﻿type
  Graph=class;
  
  GraphNode = class
    
    gr: Graph;
    
    data: string;
    next: array of GraphNode;
    
    function CostTo(next: GraphNode): integer;
    begin
      var s1 := self.data;
      var s2 := next.data;
      
      var c := Min(s1.Length, s2.Length);
      
      for var i := 1 to c do
        if s1[i]=s2[i] then c -= 1;
      
      Result := c;
    end;
    
    function FindPath(last: GraphNode; prev: HashSet<GraphNode>; max_depth: integer; done: Dictionary<GraphNode, array of GraphNode>): List<GraphNode>;
    begin
      
      if self=last then
      begin
        Result := new List<GraphNode>;
        Result += self;
        exit;
      end;
      
      if max_depth=0 then exit;//Result := nil;
      
      begin
        var res: array of GraphNode;
        if done.TryGetValue(self, res) then
        begin
          Result := res.Count-1>max_depth?nil:res.ToList;
          exit;
        end;
      end;
      
      if self.next.All(n->prev.Contains(n)) then exit;//Result := nil;
      
      //=================================================
      
      var enm: IEnumerator<GraphNode> :=
        IEnumerable&<GraphNode>(      //ToDo #1647
          self.next
          .Where(n->not prev.Contains(n))
          .GroupBy(last.CostTo)
          .MinBy(g->g.Key)
        ).GetEnumerator();
      
      //=================================================
      
      while enm.MoveNext do
      begin
        Result := enm.Current.FindPath(last, (prev+self).ToHashSet, max_depth=-1?-1:max_depth-1, done);
        if Result <> nil then
        begin
          //
          break;
        end;
      end;
      if Result=nil then exit;
      
      //=================================================
      
      max_depth := Result.Count-1;
      while enm.MoveNext do
      begin
        var curr: List<GraphNode> := enm.Current.FindPath(last, (prev+self).ToHashSet, max_depth, done);
        if curr = nil then continue;
        if curr.Count < Result.Count then
        begin
          Result := curr;
          max_depth := Result.Count-1;
        end;
      end;
      
      //=================================================
      
      Result.Insert(0, self);
      writeln($'{done.Count}: {self.data} => ', Result.Skip(1).Select(n->n.data).JoinIntoString);//дебаг
      done[self] := Result.ToArray;
    end;
    
    function FindPath(last: GraphNode) :=
    FindPath(last, new HashSet<GraphNode>, -1,  new Dictionary<GraphNode, array of GraphNode>);
    
    public function ToString: string; override :=
    $'"{data}" => ["{next.Select(n->n.data).JoinIntoString(''", "'')}"]';
    
  end;
  Graph = class
    
    nodes: array of GraphNode;
    
    constructor(br: System.IO.BinaryReader);
    begin
      nodes := new GraphNode[br.ReadInt32];
      for var i := 0 to nodes.Length-1 do nodes[i] := new GraphNode;
      
      for var i := 0 to nodes.Length-1 do
      begin
        var n := nodes[i];
        
        n.gr := self;
        n.data := br.ReadString;
        
        n.next := new GraphNode[br.ReadInt32];
        for var i2 := 0 to n.next.Length-1 do
          n.next[i2] := nodes[br.ReadInt32];
        
      end;
      
    end;
    
  end;

function LoadGraphs(fname: string := 'words.database'): array of Graph;
begin
  var br := new System.IO.BinaryReader(System.IO.File.OpenRead(fname));
  Result := new Graph[br.ReadInt32];
  for var i := 0 to Result.Length-1 do
    Result[i] := new Graph(br);
end;

type
  NodeFindingException = class(Exception)
    constructor(text: string) :=
    inherited Create(text);
  end;

begin
  var ghs := LoadGraphs;
  
  while true do
  try
    //var w1 := ReadlnString('Слово 1:').ToLower;
    //var w2 := ReadlnString('Слово 2:').ToLower;
    var w1 := 'слон';
    var w2 := 'люстра';
    
    var n1 := ghs.SelectMany(g->g.nodes).FirstOrDefault(n->n.data=w1);
    var n2 := ghs.SelectMany(g->g.nodes).FirstOrDefault(n->n.data=w2);
    
    if n1=nil then raise new NodeFindingException($'Слово "{w1}" не найдено в базе данных');
    if n2=nil then raise new NodeFindingException($'Слово "{w2}" не найдено в базе данных');
    if n1.gr <> n2.gr then raise new NodeFindingException($'Нет пути между "{w1}" и "{w2}"');
    
    var sw := new System.Diagnostics.Stopwatch;
    
    //writeln;
    sw.Restart;
    var path := n1.FindPath(n2);
    sw.Stop;
    
    writeln;
    writeln('готовый путь:');
    path
    .Select(n->n.data)
    .PrintLines;
    
    writeln;
    writeln($'Поиск занял {sw.Elapsed}');
    
    readln;
    System.Console.Clear;
  except
    on e: NodeFindingException do
    begin
      System.Console.Clear;
      writeln('Ошибка!');
      writeln(e.Message);
      writeln;
    end;
  end;
  
end.